#include <Python.h>

#define VALIDATE(x)         do { if ((x) == NULL) return NULL; } while (0)
#define CALLABLE_CHECK(x,i) \
  do { \
    if (!PyCallable_Check(x)) {                                         \
      PyErr_SetString(PyExc_TypeError, "argument " i " must be callable"); \
      return NULL; \
    } \
  } while (0)
#define OBJ2BOOL_ERROR PyErr_SetString(PyExc_ValueError, "cannot convert value to bool")
#define Py_DECREF_ARGS1(x) \
  PyTuple_SET_ITEM(x, 0, NULL); \
  Py_DECREF(x)
#define Py_DECREF_ARGS2(x) \
  PyTuple_SET_ITEM(x, 0, NULL); \
  PyTuple_SET_ITEM(x, 1, NULL); \
  Py_DECREF(x)

#define ITER_DECL \
  int _istuple, _islist, _isindexed; \
  Py_ssize_t _seq_size = 0
#define ITER_CHECK(seq)                         \
  do {                                          \
    _istuple = PyTuple_Check(seq);              \
    _islist = PyList_Check(seq);                \
    if (_istuple)                               \
      _seq_size = PyTuple_Size(seq);            \
    else if (_islist)                           \
      _seq_size = PyList_Size(seq);             \
    else                                        \
      _seq_size = PySequence_Size(seq);         \
    _isindexed = _istuple || _islist;           \
  } while (0)
#define ITER_INIT(seq, it, n, fargs) \
  ((_isindexed && !get_iter_fargs(NULL, n, NULL, &fargs)) ||    \
   !get_iter_fargs(seq, n, &it, &fargs))
#define IS_SEQ_EMPTY        (_seq_size <= 0)
#define ITER_WHILE(i)                           \
  i = -1;                                       \
  while (!IS_SEQ_EMPTY && ++i < _seq_size)
#define SET_ITER_ITEM(seq, i, x, it)          \
  if (_isindexed)                             \
    if (_istuple)                             \
      x = PyTuple_GET_ITEM(seq, i);           \
    else                                      \
      x = PyList_GET_ITEM(seq, i);            \
  else if ((x = PyIter_Next(it)) == NULL)     \
    break
#define ITER_DECREF(x) do { if (!_isindexed) { Py_DECREF(x); } } while (0)

static int get_iter_fargs(PyObject *, int, PyObject **, PyObject **);
static int find(PyObject *, PyObject *, int, int,
                PyObject **, Py_ssize_t *,
                Py_ssize_t **, Py_ssize_t, Py_ssize_t *);
static int span(PyObject *, PyObject *, PyObject **, PyObject **);

static int
get_iter_fargs(PyObject *obj, int size, PyObject **itp, PyObject **fargsp)
{
  PyObject *it = NULL, *fargs;

  if (obj)
    it = PyObject_GetIter(obj);
  fargs = PyTuple_New(size);
  if (PyErr_Occurred()) {
    Py_XDECREF(it);
    Py_XDECREF(fargs);
    return 0;
  } else {
    if (it) *itp = it;
    *fargsp = fargs;
    return 1;
  }
}

/*
 * sequence     affirm      negate
 * -------------------------------
 * empty             0           1
 * found             1           0
 * not found         0           1
 * 
 */
static int
find(PyObject *f, PyObject *seq, int negate, int all,
     PyObject **value, Py_ssize_t *index,
     Py_ssize_t **indices, Py_ssize_t default_size, Py_ssize_t *indices_size)
{
  PyObject *it = NULL, *x, *fargs, *res;
  int res_status;
  Py_ssize_t i = 0, found = 0;
  ITER_DECL;

  ITER_CHECK(seq);
  if (IS_SEQ_EMPTY)
    goto notfound;
  if (ITER_INIT(seq, it, 1, fargs))
      return -1;

  ITER_WHILE(i) {
    SET_ITER_ITEM(seq, i, x, it);
    PyTuple_SET_ITEM(fargs, 0, x);
    if ((res = PyObject_CallObject(f, fargs)) == NULL)
      goto error;
    res_status = PyObject_IsTrue(res);
    if ((res_status == 1 && !negate) ||
        (res_status == 0 && negate)) {
      ITER_DECREF(x);
      Py_DECREF(res);
      if (all) {
        if (indices != NULL) {
          if (found >= default_size)
            PyMem_Realloc(*indices, found * 2);
          (*indices)[found++] = i;
        }
      } else {
        ITER_DECREF(it);
        Py_DECREF_ARGS1(fargs);
        if (value != NULL) *value = x;
        if (index != NULL) *index = i;
        return 1;
      }
    } else if (res_status < 0) {
      OBJ2BOOL_ERROR;
      goto error;
    } else if (PyErr_Occurred()) {
        Py_XDECREF(res);
        goto error;
    } else
      ITER_DECREF(x);
  }

  ITER_DECREF(it);
  Py_DECREF_ARGS1(fargs);

  if (IS_SEQ_EMPTY)
    goto notfound;
  else if (all) {
    if (indices_size != NULL)
      *indices_size = found;
    return negate ? !found : found > 0;
  } else if (found)
    return 1;

notfound:
  PyErr_SetString(PyExc_LookupError, "not found in sequence");
  return IS_SEQ_EMPTY && negate;

error:
  ITER_DECREF(x);
  ITER_DECREF(it);
  Py_DECREF_ARGS1(fargs);
  return -1;
}


static PyObject *
hof_find(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *value;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  status = find(f, seq, 0, 0, &value, NULL, NULL, 0, NULL);
  if (status > 0) {
    Py_INCREF(value);
    return value;
  } else
    return NULL;
}

static PyObject *
hof_find_index(PyObject *self, PyObject *args)
{
  PyObject *f, *seq;
  Py_ssize_t index;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  status = find(f, seq, 0, 0, NULL, &index, NULL, 0, NULL);
  if (status > 0)
    return PyInt_FromLong(index);
  else
    return NULL;
}

static PyObject *
hof_find_indices(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *indices_list;
  Py_ssize_t *indices, indices_size = 0, default_size, i;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  if ((default_size = PySequence_Size(seq)) == -1)
    default_size = 256 / sizeof(Py_ssize_t);
  indices = PyMem_New(Py_ssize_t, default_size);

  status = find(f, seq, 0, 1, NULL, NULL,
                &indices, default_size, &indices_size);
  if (status >= 0) {
    VALIDATE(indices_list = PyList_New(indices_size));
    for (i = 0; i < indices_size; i++)
      PyList_SET_ITEM(indices_list, i, PyInt_FromLong(indices[i]));
    PyMem_Free(indices);
    return indices_list;
  } else
    return NULL;
}

static PyObject *
hof_any(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *res;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  status = find(f, seq, 0, 0, NULL, NULL, NULL, 0, NULL);
  if (status >= 0) {
    res = status > 0 ? Py_True : Py_False;
    Py_INCREF(res);
    PyErr_Clear();
    return res;
  } else
    return NULL;
}

static PyObject *
hof_all(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *res;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  status = find(f, seq, 1, 0, NULL, NULL, NULL, 0, NULL);
  if (status >= 0) {
    res = status > 0 ? Py_False : Py_True;
    Py_INCREF(res);
    PyErr_Clear();
    return res;
  } else
    return NULL;
}

static PyObject *
hof_take(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *takes;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  status = span(f, seq, &takes, NULL);
  if (status > 0)
    return takes;
  else
    return NULL;
}

static PyObject *
hof_drop(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *drops;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  status = span(f, seq, NULL, &drops);
  if (status > 0)
    return drops;
  else
    return NULL;
}

static PyObject *
hof_span(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *takes, *drops, *value;
  int status;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  status = span(f, seq, &takes, &drops);
  if (status > 0) {
    VALIDATE(value = PyTuple_New(2));
    PyTuple_SET_ITEM(value, 0, takes);
    PyTuple_SET_ITEM(value, 1, drops);
    return value;
  } else
    return NULL;
}

static int
span(PyObject *f, PyObject *seq, PyObject **takes, PyObject **drops)
{
  PyObject *it, *x, *fargs, *res = NULL,
    **tmp_takes = NULL, **tmp_drops = NULL;
  int res_status;
  Py_ssize_t size, i, j, takes_i = 0, drops_i = 0, start_drop = 0;
  ITER_DECL;

  ITER_CHECK(seq);
  if (ITER_INIT(seq, it, 1, fargs))
      return 0;

  size = PyObject_Length(seq);
  if (takes != NULL)
    if ((tmp_takes = PyMem_New(PyObject *, size)) == NULL) {
      PyErr_NoMemory();
      return 0;
    }
  if (drops != NULL) {
    if ((tmp_drops = PyMem_New(PyObject *, size)) == NULL) {
      PyErr_NoMemory();
      return 0;
    }
  }

  ITER_WHILE(i) {
    SET_ITER_ITEM(seq, i, x, it);
    if (!start_drop) {
      PyTuple_SET_ITEM(fargs, 0, x);
      if ((res = PyObject_CallObject(f, fargs)) == NULL)
        goto error;
      res_status = PyObject_IsTrue(res);
      Py_DECREF(res);
      if (res_status == 0)
        start_drop = 1;
      else if (res_status < 0) {
        OBJ2BOOL_ERROR;
        goto error;
      }
    }
    if (start_drop) {
      if (tmp_drops != NULL) {
        Py_INCREF(x);
        tmp_drops[drops_i++] = x;
      }
    } else if (tmp_takes != NULL) {
      Py_INCREF(x);
      tmp_takes[takes_i++] = x;
    }
    if (PyErr_Occurred())
      goto error;
    ITER_DECREF(x);
  }

  ITER_DECREF(it);
  PyTuple_SET_ITEM(fargs, 0, NULL);
  Py_DECREF(fargs);

  if (takes != NULL) {
    *takes = PyList_New(takes_i);
    for (j = 0; j < takes_i; j++)
      PyList_SET_ITEM(*takes, j, tmp_takes[j]);
    PyMem_Free(tmp_takes);
  }
  if (drops != NULL) {
    *drops = PyList_New(drops_i);
    for (j = 0; j < drops_i; j++)
      PyList_SET_ITEM(*drops, j, tmp_drops[j]);
    PyMem_Free(tmp_drops);
  }

  return 1;

error:
  ITER_DECREF(x);
  ITER_DECREF(it);
  Py_DECREF_ARGS1(fargs);
  return 0;
}

static PyObject *
hof_partition(PyObject *self, PyObject *args)
{
  PyObject *f, *seq, *it, *x, *fargs, *satisfied, *not_satisfied, *res;
  int res_status;
  Py_ssize_t i;
  ITER_DECL;

  if (!PyArg_ParseTuple(args, "OO", &f, &seq))
    return NULL;
  CALLABLE_CHECK(f, "1");

  ITER_CHECK(seq);
  if (ITER_INIT(seq, it, 1, fargs))
      return NULL;

  satisfied = PyList_New(0);
  not_satisfied = PyList_New(0);
  if (PyErr_Occurred()) {
    Py_XDECREF(satisfied);
    Py_XDECREF(not_satisfied);
    return NULL;
  }

  ITER_WHILE(i) {
    SET_ITER_ITEM(seq, i, x, it);
    PyTuple_SET_ITEM(fargs, 0, x);
    VALIDATE(res = PyObject_CallObject(f, fargs));
    res_status = PyObject_IsTrue(res);
    if (res_status == 1)
      PyList_Append(satisfied, x);
    else if (res_status == 0)
      PyList_Append(not_satisfied, x);
    else {
      ITER_DECREF(x);
      if (PyErr_Occurred()) {
        OBJ2BOOL_ERROR;
        ITER_DECREF(it);
        Py_DECREF(res);
        Py_DECREF(fargs);
        return NULL;
      }
    }
  }

  ITER_DECREF(it);
  PyTuple_SET_ITEM(fargs, 0, NULL);
  Py_DECREF(fargs);

  VALIDATE(res = PyTuple_New(2));
  PyTuple_SET_ITEM(res, 0, satisfied);
  PyTuple_SET_ITEM(res, 1, not_satisfied);
  return res;
}

static PyMethodDef HigherOrderFunctionsMethods[] = {
  {"find", hof_find, METH_VARARGS,
   "returns the all element in the sequence matching the function, or raise LookupError if there is no such element"},
  {"find_index", hof_find_index, METH_VARARGS,
   "returns index of the first element in the sequence matching the function, or raise LookupError if there is no such element"},
  {"find_indices", hof_find_indices, METH_VARARGS,
   "returns indices of all elements in the sequence matching the function, or raise LookupError if there is no such element"},
  {"any", hof_any, METH_VARARGS,
   "return true if any element of the sequence that satisfy the function"},
  {"all", hof_all, METH_VARARGS,
   "return true if all element of the sequence that satisfy the function"},
  {"take", hof_take, METH_VARARGS,
   "returns the longest prefix (possibly empty) of elements that satisfy the function"},
  {"drop", hof_drop, METH_VARARGS,
   "returns the suffix remaining after take(f, seq)"},
  {"span", hof_span, METH_VARARGS,
   "span(f, seq) is equivalent to (take(f, seq), drop(f, seq))"},
  {"partition",  hof_partition, METH_VARARGS,
   "returns a tuple pair of the sequences which do and do not satisfy the function"},
  {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC
inithof_c(void)
{
  (void)Py_InitModule("hof_c", HigherOrderFunctionsMethods);
}
